#!/usr/bin/perl -w
use strict;

# This is not a module because perl modules make me cry.
# It is, however, the file which contains the code to parse the user's input.

my @input;
my %quoted; # these two are a hack for quoted words
my $qc = 0;
  
sub get_command
{
  while (@input == 0)
  {
    my $in = read_line("> ");
    while ($in =~ s/"(.*?)"/quotehack$qc/)
    {
      $quoted{$qc} = $1; # yes, this should really be a hash and not an array
      $qc++;
    }
    
    push @input, split(/\./, $in) if (length($in) > 0);
    
    # if this was a valid line, take an undo snapshot and record for 'again':
    if (@input > 0)
    {
      push @{ $state{'undo'} }, take_snapshot();
      $state{'lastlastcommand'} = $state{'lastcommand'} || "";
      $state{'lastcommand'} = $input[$#input];
      $state{'lastcommand'} =~ s/quotehack([0-9]+)/$quoted{$1}/eg;
    }
  }
  
  my $c = lc(shift @input);
  $c =~ s/[^a-z0-9 ]//g;
  $c =~ s/quotehack([0-9]+)/$quoted{$1}/eg;
  return parse_command($c);
}

sub unread
{
  push @input, shift;
}

sub parse_command
{
  my $command = shift;

  # pull out stopwords:
  my @words = grep { !is_stopword($_) } split ' ', $command;

  my $verb = '';
  $verb = shift @words if (@words > 0);

  # if it's a two-word verb ('look at', etc), then the preposition in question
  # will either be the first or the last word in the list
  if (@words > 0 && is_prep($words[0]))
  {
    $verb .= ' ' . shift(@words);
  }
  elsif (@words > 0 && is_prep($words[-1]))
  {
    $verb .= ' ' . pop(@words);
  }

  $verb = check_synonym($verb);

  # if there are any objects, pull those out; objects are separated by
  # prepositions
  my @dobj;
  my $prep = '';
  while (@words > 0)
  {
    my $w = shift @words;
    if (is_prep($w))
    {
      $prep = $w;
      last;
    }
    push @dobj, $w;
  }

  my @iobj;
  while (@words > 0)
  {
    my $w = shift @words;
    if (is_prep($w))
    {
      unshift @words, $w;
      last;
    }
    push @iobj, $w;
  }

  my $error = '';
  if (@words > 0)
  {
    $error = 'syntax';
  }

  # resolve the iobj
  my $iobj = lookup_vocab(\@iobj, \$error);

  # then resolve the dobj

  # special-case use of 'all':
  if (@dobj == 1 && $dobj[0] eq 'all')
  {
    # get the list of objects this applies to -- if the thing is
    # >VERB X FROM Y, pass Y in, otherwise pass -1
    my $src = index($prep, "from") == -1 ? -1 : $iobj;
    my @objlist = all_list($verb, $src);

    if (@objlist == 0)
    {
      return ($verb, $prep, -1, -1, $command, "noall $verb");
    }
    else
    {
      # enqueue a command for every object that matched in the 'all' list:
      unread("prefixon") if (@objlist > 1);
      foreach (@objlist)
      {
        next if $_ == PC_OBJ; # never include the PC in 'all'

        my $single = $verb . " " . $state{'names'}->[$_] . " " . 
            $prep . " " . join(' ', @iobj);
        unread($single);
      }
      unread("prefixoff") if (@objlist > 1);

      # then return empty to abort this command (the one with the 'all')
      return (); 
    }
  }

  my $dobj = lookup_vocab(\@dobj, \$error);

  return ($verb, $prep, $dobj, $iobj, $command, join(' ', @dobj), 
          join(' ', @iobj), $error);
}

my %preps = map { $_ => 1 } qw(with in on under behind through to for about onto into downon upon out outof from over between inbetween at thru down up off);

sub is_prep
{
  return exists($preps{$_[0]}) ? 1 : 0;
}

my %stopwords = map { $_ => 1 } qw(the a an of);

sub is_stopword
{
  return exists($stopwords{$_[0]}) ? 1 : 0;
}

my %synonyms = (
  'z' => 'wait',
  'g' => 'again',
  'q' => 'quit',
  'hint' => 'help',
  'about' => 'help',
  'version' => 'credits',
  'get' => 'take',
  'pick up' => 'take',
  'wear' => 'take',
  'don' => 'take',
  'doff' => 'drop',
  'remove' => 'drop',
  'place' => 'put',
  'hang' => 'put',
  'drink' => 'eat',
  'l' => 'look',
  'i' => 'inventory',
  'inv' => 'inventory',
  'unlock' => 'open',
  'lock' => 'close',
  'light' => 'turn on',
  'extinguish' => 'turn off',
  'toggle' => 'switch',
  'flip' => 'switch',
  'kill' => 'attack',
  'hit' => 'attack',
  'fight' => 'attack',
  'shoot' => 'attack',
  'read' => 'examine',
  'x' => 'examine',
  'look at' => 'examine',
  'look in' => 'examine',
  'ask' => 'talk to',
  'tell' => 'talk to',
  'ahead' => 'forward',
  'forwards' => 'forward',
  'back up' => 'backward',
  'back' => 'backward',
  'retreat' => 'backward',
  'behind' => 'backward',
  'climb on' => 'climb',
  'stand on' => 'climb',
  'sit on' => 'climb',
  'u' => 'up',
  'd' => 'down',
  'n' => 'north',
  's' => 'south',
  'e' => 'east',
  'w' => 'west',
  'ne' => 'northeast',
  'se' => 'southeast',
  'nw' => 'northwest',
  'sw' => 'southwest',
);

sub check_synonym
{
  my $verb = shift;
  return $synonyms{$verb} if exists $synonyms{$verb};
  return $verb;
}

my %pronouns = map { $_ => 1 } qw(it he she him her them they);

sub is_pronoun
{
  return exists($pronouns{$_[0]}) ? 1 : 0;
}

# convert words into objects; if the words match too many objects or no 
# objects, raise an error
sub lookup_vocab
{
  my ($words, $err_ref) = @_;

  return -1 if (@$words == 0);  # empty word list is ok, I guess

  if (@$words == 1 && is_pronoun($words->[0]))
  {
    return $state{'it'} if ($state{'it'} != -1);
    $$err_ref = "nopronoun";
    return -1;
  }

  my %objs;
  
  foreach my $w (@$words)
  {
    if (exists($state{'vocab'}->{$w}))
    {
      foreach my $o (@{ $state{'vocab'}->{$w} })
      {
        $objs{$o}++;
      }
    }
    else
    {
      $$err_ref = "badvoc $w";
      return -1;
    }
  }

  my @o = keys %objs;
  if (@o > 1)
  {
    # in the case of ambiguous matches, prefer an object that matches
    # every word:
    my @best = grep { $objs{$_} == @$words } @o;
    @o = @best if (@best == 1);
  }

  if (@o == 1)
  {
    # object can only be referred to if it's in the PC's room:
    if (within($o[0], $state{'objects'}->[PC_OBJ])) 
    {
      return $o[0];
    }
    else
    {
      $$err_ref = 'cantsee ' . join(' ', @$words);
      return -1;
    }
  }
  else
  {
    $$err_ref = 'ambig ' . join(' ', @$words);
    return -1;
  }
}

# return list of relevant objects for 'all'
sub all_list
{
  my ($verb, $src) = @_;
  
  my @objlist;

  # if a source is given, ignore the usual stuff and just take all the objects
  # from there:
  if ($src != -1)
  {
    for (my $i = 0; $i <= LAST_OBJ; $i++)
    {
      if ($state{'objects'}->[$i] == $src)
      {
        push @objlist, $i;
      }
    }
  }

  # otherwise go on and get from the room and the PC's inventory:
  else
  {
    my $pcloc = $state{'objects'}->[PC_OBJ]; # room the pc's in
    # assuming it's not a verb that dislikes it, add things in the room:
    if ($verb ne 'drop' && $verb ne 'put')
    {
      for (my $i = 0; $i <= LAST_OBJ; $i++)
      {
        if ($state{'objects'}->[$i] == $pcloc)
        {
          push @objlist, $i;
        }
      }
    }
    
    # assuming it's not a verb that dislikes it, add things in the pc's 
    # inventory:
    if ($verb ne 'take')
    {
      for (my $i = 0; $i <= LAST_OBJ; $i++)
      {
        if ($state{'objects'}->[$i] == PC_OBJ)
        {
          push @objlist, $i;
        }
      }
    }
  }

  # motion verbs eliminate fixed objects
  if ($verb eq 'take' || $verb eq 'put' || $verb eq 'drop')
  {
    @objlist = grep { !has_trait($_, FIXED) } @objlist;
  }

  # always eliminate hidden objects
  @objlist = grep { !has_trait($_, HIDDEN) } @objlist;

  return @objlist;
}

1;

