#!/usr/bin/perl -w
use strict;

# This is not really a module, because perl modules make me cry.
# So it is the non-module file which contains the default verb behavior.

my @rules = (
  [ 0, \&handle_quit ],
  [ 0, \&handle_undo ],
  [ 0, \&handle_save ],
  [ 0, \&handle_restore ],
  [ 0, \&handle_restart ],
  [ 0, \&handle_score ],
  [ 0, \&handle_again ],
  [ 0, \&handle_help ],
  [ 0, \&handle_credits ],
  [ 0, \&handle_prefix_on ],
  [ 0, \&handle_prefix_off ],
  [ 0, \&print_prefix ],
  [ 0, \&set_pronouns ],
  [ 1, \&check_darkness ],
  [ 2, \&handle_look ], 
  [ 2, \&handle_inventory ],
  [ 2, \&handle_wait ],
  [ 2, \&handle_yes ],
  [ 2, \&handle_no ],
  [ 2, \&handle_examine ],
  [ 2, \&handle_take ],
  [ 2, \&handle_drop ],
  [ 2, \&handle_switch ],
  [ 2, \&handle_turnon ],
  [ 2, \&handle_turnoff ],
  [ 2, \&handle_open ],
  [ 2, \&handle_close ],
  [ 2, \&handle_put ],
  [ 2, \&handle_attack ],
  [ 2, \&handle_talk_to ],
  [ 2, \&handle_dir ],
  [ 2, \&handle_everything ],
);

sub init_default_rules
{
  $state{'rules'} = \@rules;
}

sub handle_quit
{
  my ($verb, $prep, $dobj, $iobj) = @_;
  if ($verb eq "quit" && $dobj == -1)
  {
    msg("quit");
    halt();
  }

  return CONTINUE;
}

sub handle_save
{
  my ($verb, $prep, $dobj, $iobj, $command) = @_;
  if ($verb eq "save" && $command =~ /^save(?:\s+"?(\S+?)"?)?$/)
  {
    if ($state{'cgi'})
    {
      output_text("Save/restore is not available in CGI mode, although you " .
                  "can probably use the bookmarking feature of your " .
                  "browser.\n");
      return ABORT;
    }

    my $file = defined($1) ? $1 : "";
    $file = read_line("Enter file to save to: ") if ($file eq "");
    
    if (!open(OUT, ">" . $file))
    {
      output_text("Error opening '$file': $!\n");
      return ABORT;
    }
    print OUT take_snapshot() . "\n";
    close OUT;
    msg("saved");
    return ABORT;
  }

  return CONTINUE;
}

sub handle_restore
{
  my ($verb, $prep, $dobj, $iobj, $command) = @_;
  if ($verb eq "restore" && $command =~ /^restore(?:\s+"?(\S+?)"?)?$/)
  {
    if ($state{'cgi'})
    {
      output_text("Save/restore is not available in CGI mode, although you " .
                  "can probably use the bookmarking feature of your " .
                  "browser.\n");
      return ABORT;
    }

    my $file = defined($1) ? $1 : "";
    $file = read_line("Enter file to restore from: ") if $file eq "";

    if (!open(IN, $file))
    {
      output_text("Error opening '$file': $!\n");
      return ABORT;
    }
    
    my $str = join('', <IN>); # read in the whole thing as one line
    close IN;
    $str =~ s/\s//g;
    
    if (!load_snapshot($str))
    {
      output_text("Error: save file in bad format.\n");
    }
    else
    {
      msg("restored");
      output_text("\n");
      unread("look");
    }

    return ABORT;
  }

  return CONTINUE;
}

sub handle_undo
{
  my ($verb, $prep, $dobj, $iobj) = @_;
  if ($verb eq "undo" && $dobj == -1)
  {
    if ($state{'cgi'})
    {
      output_text("Undo is not available in CGI mode, although you can " .
                  "probably use the back button on your browser.\n");
      return ABORT;
    }

    if (@{ $state{'undo'} } < 2)
    {
      msg("noundo");
      return ABORT;
    }

    # undo twice, because we have to undo the undo also
    pop(@{ $state{'undo'} });
    load_snapshot(pop @{ $state{'undo'} });
    msg("undone");
    unread("look");
    return ABORT;
  }

  return CONTINUE;
}

sub handle_restart
{
  my ($verb, $prep, $dobj, $iobj) = @_;
  if ($verb eq "restart" && $dobj == -1)
  {
    return ABORT if ($state{'cgi'}); # handled elsewhere for cgi mode

    output_text($state{'intro'});
    load_snapshot($state{'init_snapshot'});
    unread("look");
    return ABORT;
  }

  return CONTINUE;
}

sub handle_score
{
  my ($verb, $prep, $dobj, $iobj) = @_;
  if ($verb eq "score" && $dobj == -1)
  {
    msg("score");
    return ABORT;
  }

  return CONTINUE;
}

sub handle_again
{
  my ($verb, $prep, $dobj, $iobj) = @_;
  if ($verb eq "again" && $dobj == -1)
  {
    if ($state{'lastlastcommand'} eq "")
    {
      msg("noagain");
      return ABORT;
    }

    # since again is the last command, use the last last
    unread($state{'lastlastcommand'}); 
    # set things up so they can again again
    $state{'lastcommand'} = $state{'lastlastcommand'};
    return ABORT;
  }

  return CONTINUE;
}

sub handle_help
{
  my ($verb, $prep, $dobj, $iobj) = @_;
  if ($verb eq "help")
  {
    output_text($state{'title'} . " is an exciting text-adventure that " .
                "will strain your emotions to the breaking-point .. " .
                "AND BEYOND! If you get stuck, try things such as 'put', " .
                "'take', or 'examine'. Good luck!\n\n[For " .
                "information/credits about Snap itself, type 'CREDITS']\n");
    return ABORT;
  }

  return CONTINUE;
}

sub handle_credits
{
  my ($verb, $prep, $dobj, $iobj) = @_;
  if ($verb eq "credits")
  {
    my @authors = 
    (
      "William Crowther",
      "Donald Woods",
      "Roger Firth", 
      "Andrew Plotkin",
      "Emily Short",
      "Michael J\. Roberts",
      "Neil K\. Guy",
      "Stephen Granade",
      "Michael Gentry",
      "Graham Nelson",
      "Scott Adams",
    );
    my $func = sub { my $n = shift; $n =~ s/^(.*) (.*)/$2 $1/; return $n; };
    @authors = map { $_->[0] }
      sort { $a->[1] cmp $b->[1] }
      map { [$_, $func->($_) ] }
      @authors;
    
    output_text("Snap was put together by Dan Shiovitz (dbs\@cs.wisc.edu) " .
                "for 4/1/04. It took about three weeks. Thanks to Emily " .
                "Short and J\. Robinson Wheeler for moral support, and to " .
                "the authors who were swiped from: " . commalist(@authors) . 
                ".\n");
    return ABORT;
  }

  return CONTINUE;
}

# display a prefix before commands (for 'all')
sub handle_prefix_on
{
  my ($verb, $prep, $dobj, $iobj) = @_;
  if ($verb eq "prefixon")
  {
    $state{'show_prefix'} = 1;
    return ABORT;
  }

  return CONTINUE;
}

# stop displaying a prefix before commands (for 'all')
sub handle_prefix_off
{
  my ($verb, $prep, $dobj, $iobj) = @_;
  if ($verb eq "prefixoff")
  {
    $state{'show_prefix'} = 0;
    return ABORT;
  }

  return CONTINUE;
}

# if we're supposed to, display a prefix for the dobj (for 'all'):
sub print_prefix
{
  my ($verb, $prep, $dobj, $iobj) = @_;
  if ($state{'show_prefix'} && $dobj != -1)
  {
    output_text(thename($dobj) . ": ");
    # fall through to continue, in any case
  }

  return CONTINUE;
}

# set the pronouns, if there are any; this always happens, regardless of 
# darkness or whatever
sub set_pronouns
{
  my ($verb, $prep, $dobj, $iobj) = @_;
  $state{'it'} = $dobj if ($dobj != -1);
  return CONTINUE;
}

sub check_darkness
{
  my ($verb, $prep, $dobj, $iobj) = @_;
  if (!is_lit($state{'objects'}->[PC_OBJ]) &&
      (($dobj != -1 && !has_trait($dobj, LIGHT)) ||
       ($iobj != -1 && !has_trait($iobj, LIGHT))))
  {
    msg("toodark");
    return ABORT;
  }
  return CONTINUE;
}

sub handle_look
{
  my ($verb, $prep, $dobj, $iobj) = @_;
  if ($verb eq "look" && $dobj == -1)
  {
    if (!is_lit($state{'objects'}->[PC_OBJ]))
    {
      output_text("Darkness\n");
      msg("darkroom");
      return SUCCESS;
    }

    my $desc = $state{'names'}->[$state{'objects'}->[PC_OBJ]] . "\n";
    $desc .= $state{'descriptions'}->[$state{'objects'}->[PC_OBJ]] . "\n";
      
    my $cont = list_contents($state{'objects'}->[PC_OBJ]);
    if ($cont ne "")
    {
      $desc .= $cont . "\n";
    }

    output_text($desc);
    return SUCCESS;
  }
  return CONTINUE;
}

sub handle_inventory
{
  my ($verb, $prep, $dobj, $iobj, $command) = @_;
  if ($verb eq "inventory" || $command eq "take inventory")
  {
    my $inv = list_contents(PC_OBJ, 1);
    $inv = "You are empty-handed." if ($inv eq "");
    output_text($inv . "\n");

    return SUCCESS;
  }
  return CONTINUE;
}

sub handle_wait
{
  my ($verb, $prep, $dobj, $iobj) = @_;
  if ($verb eq "wait")
  {
    output_text("Time passes ...\n");
    return SUCCESS;
  }
  return CONTINUE;
}

sub handle_yes
{
  my ($verb, $prep, $dobj, $iobj) = @_;
  if ($verb eq "yes")
  {
    output_text("You seem very positive.\n");
    return SUCCESS;
  }
  return CONTINUE;
}

sub handle_no
{
  my ($verb, $prep, $dobj, $iobj) = @_;
  if ($verb eq "no")
  {
    output_text("You seem very negative.\n");
    return SUCCESS;
  }
  return CONTINUE;
}

sub handle_examine
{
  my ($verb, $prep, $dobj, $iobj) = @_;
  if (($verb eq "examine" || $verb eq "look") && $dobj != -1)
  {
    my $desc = $state{'descriptions'}->[$dobj] . " ";
    if (has_trait($dobj, SWITCHABLE))
    {
      $desc .= ucfirst(thename($dobj)) . " is " . 
          (flagged($dobj) ? "on" : "off") . ". ";
    }
    elsif (has_trait($dobj, LOCKABLE))
    {
      $desc .= ucfirst(thename($dobj)) . " is " . 
          (flagged($dobj) ? "locked" : "open") . ". ";
    }
    elsif (has_trait($dobj, OPENABLE))
    {
      $desc .= ucfirst(thename($dobj)) . " is " . 
          (flagged($dobj) ? "closed" : "open") . ". ";
    }

    $desc .= list_contents($dobj) . "\n";

    output_text($desc);

    return SUCCESS;
  }
  return CONTINUE;
}

sub handle_take
{
  my ($verb, $prep, $dobj, $iobj) = @_;
  if ($verb eq "take" && $dobj != -1)
  {
    if (within($dobj, PC_OBJ))
    {
      msg("areholding", $dobj);
      return ABORT;
    }
    elsif (has_trait($dobj, FIXED))
    {
      msg("cantmove", $dobj);
      return ABORT;
    }
    else
    {
      output_text("Taken.\n");
      $state{'objects'}->[$dobj] = PC_OBJ;
      return SUCCESS;
    }
  }
  return CONTINUE;
}

sub handle_drop
{
  my ($verb, $prep, $dobj, $iobj) = @_;
  if ($verb eq "drop" && $dobj != -1)
  {
    if (!within($dobj, PC_OBJ))
    {
      msg("arentholding", $dobj);
      return ABORT;
    }
    elsif (has_trait($dobj, FIXED))
    {
      msg("cantmove", $dobj);
      return ABORT;
    }
    else
    {
      output_text("Dropped.\n");
      $state{'objects'}->[$dobj] = $state{'objects'}->[PC_OBJ];
      return SUCCESS;
    }
  }
  return CONTINUE;
}

sub handle_switch
{
  my ($verb, $prep, $dobj, $iobj) = @_;
  if ($verb eq "switch" && $dobj != -1)
  {
    if (flagged($dobj))
    {
      return handle_turnoff("turn off", @_[1..$#_]);
    }
    else
    {
      return handle_turnon("turn on", @_[1..$#_]);
    }
  }
  return CONTINUE;
}

sub handle_turnon
{
  my ($verb, $prep, $dobj, $iobj) = @_;
  if ($verb eq "turn on" && $dobj != -1)
  {
    if (!has_trait($dobj, SWITCHABLE))
    {
      output_text("That can't be turned on.\n");
      return ABORT;
    }
    else
    {
      output_text("You turn on " . thename($dobj) . ".\n");
      flagged($dobj, 1);
      return SUCCESS;
    }
  }
  return CONTINUE;
}

sub handle_turnoff
{
  my ($verb, $prep, $dobj, $iobj) = @_;
  if ($verb eq "turn off" && $dobj != -1)
  {
    if (!has_trait($dobj, SWITCHABLE))
    {
      output_text("That can't be turned off.\n");
      return ABORT;
    }
    else
    {
      output_text("You turn off " . thename($dobj) . ".\n");
      flagged($dobj, 0);
      return SUCCESS;
    }
  }
  return CONTINUE;
}

sub handle_open
{
  my ($verb, $prep, $dobj, $iobj) = @_;
  if ($verb eq "open" && $dobj != -1)
  {
    if (!has_trait($dobj, OPENABLE))
    {
      output_text("That isn't openable.\n");
      return ABORT;
    }
    elsif (!flagged($dobj))
    {
      output_text(ucfirst(thename($dobj)) . " is already open.\n");
      return ABORT;
    }
    elsif (!has_trait($dobj, LOCKABLE)) # if it's openable but not lockable
    {
      output_text("You open " . thename($dobj) . ".\n");
      flagged($dobj, 0);
      return SUCCESS;
    }
    elsif ($iobj == -1 || $prep ne "with")
    {
      output_text("You must specify what to unlock " . thename($dobj) . 
                  " with.\n");
      return ABORT;
    }
    if (!has_trait($iobj, KEY))
    {
      output_text("You can't unlock anything with " . thename($iobj) . ".\n");
      return ABORT;
    }
    else
    {
      output_text("You unlock and open " . thename($dobj) . ".\n");
      flagged($dobj, 0);
      return SUCCESS;
    }
  }
  return CONTINUE;
}

sub handle_close
{
  my ($verb, $prep, $dobj, $iobj) = @_;
  if ($verb eq "close" && $dobj != -1)
  {
    if (!has_trait($dobj, OPENABLE))  # if it's not openable
    {
      output_text("That isn't closable.\n");
      return ABORT;
    }
    elsif (flagged($dobj))
    {
      output_text(ucfirst(thename($dobj)) . " is already closed.\n");
      return ABORT;
    }
    else
    {
      output_text("You close " . 
                  (has_trait($dobj, LOCKABLE) ? "and lock " : "") . 
                  thename($dobj) . ".\n");
      flagged($dobj, 1);
      return SUCCESS;
    }
  }
  return CONTINUE;
}

sub handle_put
{
  my ($verb, $prep, $dobj, $iobj) = @_;
  if ($verb eq "put" && $dobj != -1 && $iobj != -1)
  {
    if (!within($dobj, PC_OBJ))
    {
      msg("arentholding", $dobj);
      return ABORT;
    }
    elsif (has_trait($dobj, FIXED))
    {
      msg("cantmove", $dobj);
      return ABORT;
    }
    elsif ($prep eq "in")
    {
      if (has_trait($iobj, CONTAINER))
      {
        if (flagged($iobj) || !has_trait($iobj, OPENABLE))
        {
          output_text("You put " . thename($dobj) . " in " . thename($iobj) . 
                      ".\n");
          $state{'objects'}->[$dobj] = $iobj;
          return SUCCESS;
        }
        else
        {
          output_text(ucfirst(thename($iobj)) . " is closed.\n");
          return ABORT;
        }
      }
      else
      {
        output_text("You can't put anything in " . thename($iobj) . ".\n");
        return ABORT;
      }
    }
    elsif ($prep eq "on")
    {
      if (has_trait($iobj, SURFACE))
      {
        output_text("You put " . thename($dobj) . " on " . thename($iobj) . 
                    ".\n");
        $state{'objects'}->[$dobj] = $iobj;
        return SUCCESS;
      }
      else
      {
        output_text("You can't put anything on " . thename($iobj) . ".\n");
        return ABORT;
      }
    }

    # otherwise fallthrough to the return CONTINUE, since we didn't
    # understand this
  }
  return CONTINUE;
}

sub handle_attack
{
  my ($verb, $prep, $dobj, $iobj) = @_;
  if ($verb eq "attack" && $dobj != -1)
  {
    msg("noviolence");
    return ABORT;
  }

  return CONTINUE;
}

sub handle_talk_to
{
  my ($verb, $prep, $dobj, $iobj) = @_;
  if ($verb eq "talk to" && $dobj != -1)
  {
    if (has_trait($dobj, ACTOR))
    {
      output_text("\"I don't know much about that,\" " . thename($dobj) . 
                  " says.\n");
      return SUCCESS;
    }
    else
    {
      output_text("You can't talk to that.\n");
      return ABORT;
    }
  }

  return CONTINUE;
}

BEGIN
{
  my %dirs = map { $_ => 1 } qw(north south east west northeast southeast 
                                northwest southwest up down in out);
  sub is_dir
  {
    my $word = shift;
    return (exists $dirs{$word}) ? 1 : 0;
  }
}

sub handle_dir
{
  my ($verb, $prep, $dobj, $iobj, $command, $dobj_str, $iobj_str, $error) = @_;
  
  if (is_dir($verb))
  {
    my $room = $state{'objects'}->[PC_OBJ];
    if (exists $state{'objects'}->[$room]->{$verb})
    {
      $state{'objects'}->[PC_OBJ] = $state{'objects'}->[$room]->{$verb};
      handle_look("look", "", -1, -1, $command, "", "", $error);
      return SUCCESS;
    }
    else
    {
      output_text("You can't go that way.\n");
      return ABORT;
    }
  }
  return CONTINUE;
}

sub handle_everything
{
  my ($verb, $prep, $dobj, $iobj, $command, $dobj_str, $iobj_str, $error) = @_;
  
  if ($error eq "")
  {
    msg("syntax");
  }
  else
  {
    my ($err, $args) = split ' ', $error, 2;
    msg($err, $args);
  }

  return ABORT;
}


# recursively list the contents of an object; if this is a recursive call,
# list the contents more briefly and print the object's name first
sub list_contents
{
  my ($obj, $with_flags, $recursing) = @_;
  $with_flags = 0 unless defined $with_flags;
  $recursing = 0 unless defined $recursing;

  my $ret = "";

  if ($recursing)
  {
    $ret = aname($obj);

    $ret = "*" . $ret . "*" 
        if ($state{'theme'} eq "scott" && has_trait($obj, TREASURE));

    if (has_trait($obj, WEARABLE) && within($obj, PC_OBJ))
    {
      $ret .= " (being worn)";
    }
    if (has_trait($obj, LIGHT) && flagged($obj))
    {
      $ret .= " (providing light)";
    }
  }

  # if this is a closed container, then bug out at this point
  if (has_trait($obj, CONTAINER) && has_trait($obj, OPENABLE) && flagged($obj))
  {
    return $ret;
  }
  
  # collect all the contents (except the PC)
  my @objs = grep { $_ ne PC_OBJ } contents($obj);

  # sort out fixed from non-fixed
  my @fixed = grep { has_trait($_, FIXED) } @objs;
  @objs = grep { !(has_trait($_, FIXED) || has_trait($_, HIDDEN)) } @objs;

  if (@objs > 0) # skip normal listing if no objs
  {
    if ($obj eq PC_OBJ)
    {
      $ret = "You are carrying ";
    }
    elsif ($obj > LAST_OBJ)
    {
      $ret = "You see here ";
    }
    else
    {
      if (has_trait($obj, SURFACE))
      {
        if ($recursing)
        {
          $ret .= " (on which ";
        }
        else
        {
          $ret = "Resting on " . thename($obj) . " ";
        }
      }
      else
      {
        if ($recursing)
        {
          $ret .= " (in which ";
        }
        else
        {
          $ret = "Inside " . thename($obj) . " ";
        }
      }
      
      if (@objs == 1)
      {
        $ret .= "is ";
      }
      else
      {
        $ret .= "are ";
      }
    }
    
    $ret .= commalist(map { list_contents($_, $with_flags, 1) } @objs);
    if ($recursing)
    {
      $ret .= ")";
    }
    else
    {
      $ret .= ". ";
    }
  }

  # now list the contents of fixed objects (there's no way to do it nicely
  # if we're recursing, so ignore that case)
  if (@fixed > 0 && !$recursing)
  {
    # don't list the fixed objects themselves, but show their contents, if
    # they have any:
    foreach my $f (@fixed)
    {
      $ret .= list_contents($f, $with_flags, $recursing);
    }
  }

  return $ret;
}

1;

