#!/usr/bin/perl -w

##########################################################################
# Snap! 
# The exciting new IF authoring system by Dan Shiovitz (dbs@cs.wisc.edu)
##########################################################################

use strict;
use Text::Wrap;
use Carp qw(confess);
use CGI qw/:standard -debug/;

use vars qw(%state);

BEGIN
{
  require 'Constants.pm';
  require 'Parser.pm';
  require 'Verbs.pm';
  require 'World.pm';
}

my ($source, $snapshot, $cgi_mode);

if (exists $ENV{'REQUEST_METHOD'})
{
  $cgi_mode = 1;
}
elsif (grep { lc($_) eq "-cgi" } @ARGV)
{
  $cgi_mode = 1;
  @ARGV = grep { lc($_) eq "-cgi" } @ARGV;
} 

if (!$cgi_mode)
{
  my ($gamefile, $savefile);
  parse_args(\$gamefile, \$savefile);
  $source = read_source($gamefile);
  $snapshot = read_savefile($savefile);
}
else
{
  $state{'cgi'} = new CGI();
  $source = clean_source($state{'cgi'}->param("source") || "");
  $snapshot = ($state{'cgi'}->param("snapshot") || "");
  $state{'input'} = $state{'cgi'}->param("input") || "";
  $snapshot = "" if (lc($state{'input'}) eq "restart"); # hack for restart
  # another hack to show the room desc each turn
  unread("look") if (lc($state{'input'}) ne "look" && $state{'input'} ne "");
}

$state{'source'} = $source;

if ($source =~ /^[hq9+]+$/)
{
  run_hq9($source);
}
else
{
  $source = "_" x 10 if ($source eq ""); # special bonus game
  init_game($source, $snapshot);
  run_game($snapshot eq "");
}

exit 0;

######################################################################

sub parse_args
{
  my ($gref, $sref) = @_;

  my $thisname = $0;
  $thisname =~ s|.*/||;
  my $syntax = "Syntax: $thisname [-r savefile-to-restore] gamefile\n";
  
  die $syntax unless @ARGV;
  
  while (1)
  {
    my $arg = shift @ARGV;
    if (substr($arg, 0, 1) eq "-")
    {
      if ($arg eq "-r")
      {
        die $syntax unless @ARGV;
        $$sref = shift @ARGV;
        die "'$$sref': No such file\n" unless -f $$sref; # ensure file exists
      }
      else
      {
        die "Unknown option $arg; $syntax";
      }
    }
    else
    {
      $$gref = $arg;
      die "'$$gref': No such file\n" unless -f $$gref; # ensure file exists
      return;
    }
  }
}

sub read_source
{
  my $file = shift;
  open(IN, $file) or die "Can't open '$file': $!\n";
  my $str = join('', <IN>); # read in the whole thing as one line
  close IN or die "Can't close '$file': $!\n";

  return clean_source($str);
}

# fix up a string of source code:
sub clean_source
{  
  my $str = shift;

  $str = lc($str);
  $str =~ s/\s//g;
  
  # all valid hq9+ programs are valid programs:
  return $str if $str =~ /^[hq9+]+$/;
  
  # otherwise strip misc chars:
  $str =~ s/[^a-z]//g;

  # and hq, since those are reserved
  $str =~ s/[hq]//g;
  
  # if longer than ten characters return only the first ten:
  if (length($str) >= 10)
  {
    return substr($str, 0, 10);
  }

  # otherwise, pad it out to 10 characters by repeating the last character, 
  # unless it's the empty string:
  elsif (length($str) == 0)
  {
    return $str;
  }
  else
  {
    return $str . substr($str, -1, 1) x (10 - length($str));
  }
}

sub read_savefile
{
  my $file = shift;
  return "" unless defined($file) && -f $file;

  open(IN, $file) or die "Can't open '$file': $!\n";
  my $str = join('', <IN>); # read in the whole thing as one line
  close IN or die "Can't close '$file': $!\n";
  $str =~ s/\s//g;
  return $str;
}

# Interpret source as HQ9+ (or HQ9++, perhaps)
sub run_hq9
{
  my $prog = shift;
  my $accumulator = 0;
  $state{'theme'} = 'hq9';

  foreach my $c (split //, $prog)
  {
    if ($c eq "h")
    {
      output_text("Hello, world!\n");
    }
    elsif ($c eq "q")
    {
      output_text($prog . "\n");
    }
    elsif ($c eq "9")
    {
      for (my $i = 99; $i > 1; $i--)
      {
        output_text("$i bottles of beer on the wall,\n$i bottles of beer,\n" .
                    "Take one down, pass it around,\n");
        output_text(($i-1 == 1) ? "1 bottle" : $i-1 . " bottles");
        output_text(" of beer on the wall.\n\n");
      }

      output_text("One bottle of beer on the wall,\nOne bottle of beer,\n" .
                  "Take it down, pass it around,\nNo bottles of beer on " .
                  "the wall.\n\n");
    }
    elsif ($c eq "+")
    {
      $accumulator++; # you didn't think we would *cheat*, did you?
    }
    else
    {
      halt("Serious error in HQ9+ interpreter! Unknown char '$c'\n");
    }
  }

  halt();
}

sub init_game
{
  my ($source, $snapshot) = @_;

  my @chars = split //, $source;

  my $theme = get_theme($chars[0]);
  $state{'theme'} = $theme;
  $state{'intro'} = $intros{$theme};
  init_msgs(%{ $parser_msgs{$theme} });

  # make the pc object specially:
  $state{'names'}->[PC_OBJ] = "adventurer";
  $state{'descriptions'}->[PC_OBJ] = lookup_msg('pcdesc');
  $state{'objects'}->[PC_OBJ] = get_start_room($chars[0]);
  give_trait(PC_OBJ, ACTOR);
  give_trait(PC_OBJ, FIXED);

  # add in pc vocab
  $state{'vocab'}->{'me'} = [ PC_OBJ ];
  $state{'vocab'}->{'myself'} = [ PC_OBJ ];
  $state{'vocab'}->{'adventurer'} = [ PC_OBJ ];
  $state{'vocab'}->{'cretin'} = [ PC_OBJ ];     # (obviously not *that* PC)

  # now load the objects, game objects before game rooms:
  $state{'loadidx'} = 1; # id of first loaded item

  my @objs = ( 
    $objects{get_world($chars[4])}->[0],
    $objects{get_world($chars[7])}->[1],
    $objects{get_world($chars[1])}->[2],
    $objects{get_world($chars[4])}->[3],
    $objects{get_world($chars[7])}->[4],
    $objects{get_world($chars[1])}->[5],
    $objects{get_world($chars[4])}->[6],
    $objects{get_world($chars[7])}->[7],
    $objects{get_world($chars[1])}->[8],
    $rooms{get_world($chars[2])}->[0],
    $rooms{get_world($chars[5])}->[1],
    $rooms{get_world($chars[8])}->[2],
    $rooms{get_world($chars[2])}->[3],
    $rooms{get_world($chars[5])}->[4],
    $rooms{get_world($chars[8])}->[5],
    $rooms{get_world($chars[2])}->[6],
    $rooms{get_world($chars[5])}->[7],
    $rooms{get_world($chars[8])}->[8],
  );
  init_objects(@objs);

  # last make the nowhere room: (for deleted objects)
  $state{'names'}->[LAST_ROOM] = "Nowhere";
  $state{'descriptions'}->[LAST_ROOM] = "This is nowhere.";
  $state{'objects'}->[LAST_ROOM] = {};

  # and the PC's inventory:
  foreach (@{ get_inventory($chars[0]) })
  {
    $state{'objects'}->[$_] = PC_OBJ;
  }

  # load the rules in
  init_default_rules();

  init_rules(@{ $rules{get_world($chars[3])}->[0] });
  init_rules(@{ $rules{get_world($chars[6])}->[1] });
  init_rules(@{ $rules{get_world($chars[9])}->[2] });

  # set the title based on the rooms:
  $state{'title'} = join(' ', (grep { $_ ne "" }
    get_title($chars[2])->[0], 
    get_title($chars[5])->[1], 
    get_title($chars[8])->[2]));

  # last, if there's a savefile, apply it:
  if ($snapshot ne "" && !load_snapshot($snapshot))
  {
    output_text("Error: snapshot in bad format.\n");
    halt();
  }
}

sub run_game
{
  my $new_game = shift;

  if ($new_game)
  {
    output_text("Snap v0.0001 by Dan Shiovitz\n\n" . $state{'intro'} . "\n\n");
    unread("look");
  }
 
  if ($new_game || $state{'cgi'})
  {
    $state{'init_snapshot'} = take_snapshot();
    $state{'it'} = -1; # clear pronouns
    $state{'undo'} = []; # set up undo stack
  }

  while (1)
  {
    my @parsed = get_command();
    next if @parsed == 0;

    # if we got here, the normal rule-handling kicks in; 
    # there are rules that run before stuff happens, rules that run to make
    # stuff happen, and rules that run after stuff has happened. 
    # rules have an associated stage, and the rule returns how much to
    # increment the stage by.
    
    my $stage = 0;
    foreach my $rule (@{$state{'rules'}})
    {
      # always bump the stage ahead if we're out of rules at this stage
      $stage = $rule->[0] if ($rule->[0] > $stage);

      # if this rule applies at this stage, run it
      if ($rule->[0] == $stage)
      {
        my $ret = $rule->[1]->(@parsed);
        if ($ret == SUCCESS)
        {
          $stage++; # go onto the rules at the next stage
        }
        elsif ($ret == ABORT)
        {
          last;     # stop running rules
        }
        else
        {
          ;         # just go onto the next rule
        }
      }
    }
  }
}

######################################################################
# World-model util functions:
######################################################################

# return true if $obj is (in)directly in $loc
sub within
{
  my ($obj, $loc) = @_;
  my $parent = $state{'objects'}->[$obj];
  if ($obj == $loc || $parent == $loc)
  {
    return 1;
  }
  elsif ($parent > LAST_OBJ)
  {
    return 0;
  }
  else
  {
    return within($parent, $loc);
  }
}

sub loc
{
  my ($obj, $loc) = @_;
  if (defined($loc))
  {
    $state{'objects'}->[$obj] = $loc;
  }
  else
  {
    return $state{'objects'}->[$obj];
  }
}

sub flagged
{
  my ($obj, $set) = @_;
  if (defined($set))
  {
    $state{'flags'}->[$obj] = $set;
  }
  else
  {
    return $state{'flags'}->[$obj];
  }
}

sub any_with_trait
{
  my ($obj, $trait) = @_;
  foreach my $o (contents($obj))
  {
    return $o if (has_trait($o, $trait))
  }
  return -1;
}

sub is_lit
{
  my $room = shift;
  return 1 unless has_trait($room, DARK);

  # if there's an active lightsource in the room, it's not dark
  for (my $i = 0; $i <= LAST_OBJ; $i++)
  {
    return 1 if (within($state{'objects'}->[$i], $room) && 
                 has_trait($i, LIGHT) &&
                 (!has_trait($i, SWITCHABLE) || flagged($i)));
  }

  return 0;
}

# see if an object/room has a particular trait set (returns 0 for obj=-1)
sub has_trait
{
  my ($obj, $trait) = @_;
  return (exists $state{'traits'}->[$trait]->{$obj}) ? 1 : 0;
}

# set an trait on an object/room. note that this should only be called
# during startup, because traits are not saved as part of the snapshot
sub give_trait
{
  my ($obj, $trait) = @_;
  $state{'traits'}->[$trait]->{$obj} = 1;
}

sub contents
{
  my $obj = shift;
  my @contents;
  for (my $i = 0; $i <= LAST_OBJ; $i++)
  {
    push @contents, $i if ($state{'objects'}->[$i] eq $obj);
  }
  return @contents;
}

sub aname
{
  my $obj = shift;
  my $name = $state{'names'}->[$obj];
  return $name if ($name =~ /^[A-Z]/ || $state{'theme'} eq "scott");
  return "an " . $name if ($name =~ /^[aeiou]/);
  return "a " . $name;
}

sub thename
{
  my $obj = shift;
  my $name = $state{'names'}->[$obj];
  return $name if ($name =~ /^[A-Z]/ || $state{'theme'} eq "scott");
  return "the " . $name;
}

sub commalist
{
  my @objs = @_;
  return "" if (@objs == 0);
  return $objs[0] if (@objs == 1);
  return join(" - ", @objs) if ($state{'theme'} eq "scott");
  return ($objs[0] . " and " . $objs[1]) if (@objs == 2);
  return join(", ", @objs[0..$#objs-1]) . ", and " . $objs[-1];
}

# load an undo record/savefile -- it's a list of locations, and the
# nth object is in the nth location. each item is prefixed with a ! or a .,
# to show if the nth object is flagged or unflagged.
sub load_snapshot
{
  my @chars = split //, shift();
  return 0 if $#chars > LAST_OBJ;

  my (@flags, @objs);
  
  for (my $i = 0; $i < @chars; $i++)
  {
    my $oc = ord($chars[$i]);

    $flags[$i] = 0;
    $objs[$i] = NOWHERE;

    if ($oc >= ord("A") && $oc <= ord("Z"))
    {
      $flags[$i] = 1;
      $oc -= ord("A");
    }
    else
    {
      $oc -= ord("a");
    }

    return 0 unless ($oc >= 0 && $oc <= LAST_ROOM);
    $objs[$i] = $oc;
  }
  
  $state{'flags'} = \@flags;
  splice(@{ $state{'objects'} }, 0, @objs, @objs); # replace objs but not rooms

  return 1; # savefile was all good
}

# return a snapshot of the current game state suitable for restoration
# with load_snapshot, and in url-safe ascii chars for passing around
sub take_snapshot
{
  my $str = "";
  for (my $i = 0; $i <= LAST_OBJ; $i++)
  {
    $str .= chr(ord(flagged($i) ? 'A' : 'a') + $state{'objects'}->[$i]);
  }
  return $str;
}

sub add_rule
{
  my ($priority, $rule) = @_;
  for (my $i = 0; $i < @{ $state{'rules'} }; $i++)
  {
    if ($priority <= $state{'rules'}->[$i]->[0])
    { 
      splice(@{ $state{'rules'} }, $i, 0, [$priority, $rule]);
      return;
    }
  }

  push @{ $state{'rules'} }, [$priority, $rule];
}

######################################################################
# Game responses
######################################################################

BEGIN
{
  my %msgs = (
    'win' => "\n*** You have won ***",
    'die' => "\n*** You have died ***",
    'quit' => "Thanks for playing!",
    'saved' => "Saved.",
    'restored' => "Restored.",
    'noundo' => "No further undo information is available.",
    'noall' => "There is nothing at all to %s.",
    'undone' => "Undone.",
    'score' => "Life doesn't work that way.",
    'noagain' => "No again information is available.",
    'toodark' => "It's too dark to see.",
    'darkroom' => "It is very dark here, possibly because there are no " .
              "lights.",
    'areholding' => "You are already holding %t.",
    'arentholding' => "You aren't holding %t.",
    'cantmove' => "You can't move %t.",
    'noviolence' => "Violence doesn't solve anything.",
    'syntax' => "I didn't understand that sentence.",
    'nopronoun' => "Pronouns are not currently set.",
    'badvoc' => "I don't know the word '%s'.",
    'cantsee' => "You can't see any '%s' here.",
    'ambig' => "'%s' refers to multiple objects.",
    'pcdesc' => "As good-looking as ever.",
  );

  sub lookup_msg
  {
    my ($id, $args) = @_;
    my $text;

    if (defined($args) && $args ne "")
    {
      $text = $msgs{$id};
      $args = thename($args) if ($args =~ /^\d+$/ && $text =~ s/%t/%s/);
      $args = aname($args)   if ($args =~ /^\d+$/ && $text =~ s/%a/%s/);
      $text = sprintf($text, $args) if ($text =~ /%s/);
    }
    else
    {
      $text = $msgs{$id};
    }
  }

  sub msg
  {
    output_text(lookup_msg(@_) . "\n");
  }

  sub init_msgs
  {
    my %extra = @_;
    while (my ($k, $v) = each %extra)
    {
      halt("Bad msg id '$k'\n") unless exists $msgs{$k};
      $msgs{$k} = $v;
    }
  }
}

BEGIN
{
  my $buffer = "";

  sub output_text
  {
    my $text = shift;
    $text = uc($text) if $state{'theme'} eq "scott";
    $text = piratify($text) if $state{'theme'} eq "pirate";
    $text = $state{'cgi'}->escapeHTML($text) if $state{'cgi'};
    $buffer .= $text;
  }

  sub flush_output
  {
    if (!$state{'cgi'})
    {
      # neither wrap nor fill seems to respect multiple \ns to form a paragraph
      # break, as far as I can tell. so we hack it:
      my @paras = map { wrap("", "", $_) } split /\n{2,}/, $buffer;
      print join("\n\n", @paras) . "\n";
    }
    else
    {
      $buffer =~ s/\n{2,}/<P>/g;
      $buffer =~ s/\n/<BR>/g;
      print $buffer;
    }

    $buffer = "";
  }

  sub read_line
  {
    my $prompt = shift;
    my $read = "";
    
    if (!$state{'cgi'})
    {
      flush_output();
      
      while ($read eq "")
      {
        print $prompt; 
        $read = <STDIN>;
        $read = "" unless defined $read;
        chomp $read;
      }
      
      return $read;
    }
    else
    {
      # if the user typed in some input (in which case it'd be in the url 
      # param) use that, and print their input. if not, print a form to enter
      # input and quit.
      my $input = $state{'input'} || "";
      $input =~ s/^\s*(.*?)\s*$/$1/;
      if ($input ne "")
      {
        # we can't use output_text here because we don't want this html
        # to get escaped:
        $buffer .= "\n\n<B>";
        output_text(uc($prompt) . uc($input));
        $buffer .= "</B>\n\n";
        $state{'input'} = ""; # we've now used this line, so
        return $input;
      }
      else
      {
        # we can't use output_text here because we don't want this html
        # to get escaped:
        $buffer .= 
            # start_form seems to handle urls that end in %0A very badly;
            # I can't tell what the deal is, but this seems to be a possible
            # real issue if it's taking input from a textarea, so:
            # $state{'cgi'}->start_form("GET") . 
            "<form method=\"get\" action=\"/~dans/cgi-bin/snap/snap.pl\">" .
            $state{'cgi'}->escapeHTML($prompt) . 
            $state{'cgi'}->textfield(-name => "input", -size => 50,
                                     -default => "", -override => 1) .
            $state{'cgi'}->hidden(-name => "source", 
                                  -value => $state{'source'},
                                  -override => 1) .
            $state{'cgi'}->hidden(-name => "snapshot", 
                                  -value => take_snapshot(), -override => 1) .
            $state{'cgi'}->submit("", "Enter") . 
            $state{'cgi'}->end_form;
        halt();
      }
    }
  }
}

sub piratify
{
  my $text = shift;
  $text =~ s/\b([tT])he\b/$1h'/ig;
  $text =~ s/\b([oO])f\b/$1'/ig;
  $text =~ s/\b([dD])o /$1'/ig;
  $text =~ s/\b([tT])o /$1'/ig;
  $text =~ s/\b([fF])or\b/$1er/ig;
  $text =~ s/\b([yY])ou\b/$1e/ig;
  $text =~ s/\b([yY])our\b/$1er/ig;
  $text =~ s/\b(is|are)\b/be/ig;
  $text =~ s/\b(Is|Are)\b/Be/ig;
  $text =~ s/\b(who|where|what|there|he|she|it)'s\b/$1 be/ig;
  $text =~ s/'re\b/ be/ig;
  $text =~ s/(^|\.\s+)([A-Z])/$1Arr, \l$2/ if int(rand(100)) < 10;
  return $text;
}

sub halt
{
  my $text = shift;
  output_text($text) if defined $text;
  print_page_header() if $state{'cgi'};
  flush_output();
  print_page_footer() if $state{'cgi'};
  exit(0);
}

sub print_page_header
{
  print $state{'cgi'}->header;
  print <<END;
<HTML><HEAD><TITLE>Snap Interpreter</TITLE></HEAD>
<BODY onLoad="if (document.forms.length > 0) { document.forms[0].input.focus() }">
END
}

sub print_page_footer
{
  print <<END;
  <HR>
  <A HREF="http://www.drizzle.com/~dans/">Dan Shiovitz</A> ** <A HREF="http://www.drizzle.com/~dans/if/snap/">Back to the Snap Page</A> 
  </BODY></HTML>
END
}

######################################################################
# Conditional room/object/rule data
######################################################################

sub init_objects
{
  foreach my $obj (@_)
  {
    my $idx = $state{'loadidx'}++;

    halt("error at idx $idx") unless (ref $obj eq "ARRAY" && @$obj eq 4);

    $state{'names'}->[$idx] = $obj->[0];
    $state{'descriptions'}->[$idx] = $obj->[1];
    $state{'objects'}->[$idx] = $obj->[2];
    foreach my $t (@{ $obj->[3] })
    {
      give_trait($idx, $t);

      # anything which is lockable or openable begins locked/closed:
      if ($t eq LOCKABLE || $t eq OPENABLE)
      {
        flagged($idx, 1);
      }
    }

    if ($idx <= LAST_OBJ)
    {
      my $name = lc($obj->[0]);
      $name =~ s/[^a-z0-9 ]//;

      foreach my $w (split ' ', $name)
      {
        next if is_stopword($w) || is_prep($w);
        $state{'vocab'}->{$w} = [] unless exists $state{'vocab'}->{$w};
        push @{ $state{'vocab'}->{$w} }, $idx;
      }
    }
  }
}

sub init_rules
{
  # add the rules; do it in reverse order, which will keep them in the order
  # they were originally put in (since add_rule adds each one to the front
  # of the list)
  foreach my $rule (reverse @_)
  {
    add_rule($rule->[0], $rule->[1]);
  }
}
